#include <stdint.h>
#include <string.h>
#include <sys/uio.h>
#include <terminal.h>
#include <multiboot.h>
#include <musl-glue.h>
#include <debug.h>

#define FRAMEBUFFER 0x1
#define EGA_TEXT 0x2

#define VGA_ROW(pos) ((pos)/VGA_COLS)
#define VGA_COL(pos) ((pos)%VGA_COLS)
#define VGA_POS(row, col) ((row)*VGA_COLS + (col))

static struct vga_cell buffer[VGA_SIZE];
unsigned int cursor = 0;
uint8_t format = 0x7;

static int terminal_type;

static void scroll() {
  while(cursor >= VGA_SIZE) {
    memmove(buffer,
        &buffer[VGA_POS(1,0)],
        VGA_COLS*(VGA_ROWS-1)*sizeof(struct vga_cell)
    );
    memset(&buffer[VGA_POS(VGA_ROWS-1, 0)],
        0,
        VGA_COLS*sizeof(struct vga_cell)
    );
    cursor -= VGA_COLS;
  }
}

static void terminal_write(char c) {
  int doflush = 0;
  switch(c) {
    case '\n':
      cursor += VGA_COLS - VGA_COL(cursor);
      doflush = 1;
      break;
    default:
      buffer[cursor++] = (struct vga_cell){.c = c, .f = format};
  }
  scroll();
  switch(terminal_type) {
    case FRAMEBUFFER:
      if(doflush)
        fbterm_flush(buffer);
      fbterm_movecursor(cursor);
      break;
    case EGA_TEXT:
      if(doflush)
        vga_flush(buffer);
      vga_movecursor(cursor);
      break;
  }
}

static void terminal_putsn(char *s, size_t n) {
  while(n--)
    terminal_write(*s++);
}

long k_ioctl(long fd, long cmd, long arg3, long, long, long) {
  if(fd == 1 && cmd == 0x5413) {
    struct {
      unsigned short ws_row;
      unsigned short ws_col;
      unsigned short ws_xp;
      unsigned short ws_yp;
    } *wsz = (void *)arg3;
    wsz->ws_row = 24;
    wsz->ws_col = 80;
    return 0;
  }
  long retval = -1;
  PANIC("Unknown IOCTL request: fd: %d, command: %x\n", fd, cmd);
  return retval;
}

long k_writev(long fd, long iov, long iovcnt, long, long, long) {
  if(fd == 1) {
    size_t len = 0;
    struct iovec *v = (void *) iov;
    for(int i = 0; i < iovcnt; i++) {
      terminal_putsn(v[i].iov_base, v[i].iov_len);
      len += v[i].iov_len;
    }
    return len;
  }
  long retval = 0;
  PANIC("Unknown writev request: fd: %d\n", fd);
  return retval;
}

void terminal_init() {
  struct fbinfo *fbinfo = kernel_boot_data.fbinfo;
  terminal_type = fbinfo->framebuffer_type;

  switch(terminal_type) {
    case FRAMEBUFFER:
      fbterm_init(fbinfo);
      break;
    case EGA_TEXT:
      vga_init(fbinfo);
      break;
  }
}